#!/usr/bin/python3
# encoding: utf-8
# SPDX-FileCopyrightText: 2024 FC (Fay) Stegerman <flx@obfusk.net>
# SPDX-License-Identifier: AGPL-3.0-or-later

import os
import sys
import zipfile

from typing import Optional, Tuple

from . import binres as _binres
from . import diff_zip_meta as _diff_zip_meta
from . import dump_arsc as _dump_arsc
from . import dump_axml as _dump_axml
from . import dump_baseline as _dump_baseline
from . import fix_compresslevel as _fix_compresslevel
from . import fix_files as _fix_files
from . import fix_newlines as _fix_newlines
from . import fix_pg_map_id as _fix_pg_map_id
from . import list_compresslevel as _list_compresslevel
from . import rm_files as _rm_files
from . import sort_apk as _sort_apk
from . import sort_baseline as _sort_baseline
from . import zipalign as _zipalign
from . import zipalignment as _zipalignment
from . import zipinfo as _zipinfo

import click

__version__ = "0.3.0"
NAME = "repro-apk"

ERRORS = (
    _binres.Error,
    _diff_zip_meta.Error,
    _dump_arsc.Error,
    _dump_axml.Error,
    _dump_baseline.Error,
    _fix_compresslevel.Error,
    _fix_files.Error,
    _fix_newlines.Error,
    _fix_pg_map_id.Error,
    _list_compresslevel.Error,
    _rm_files.Error,
    _sort_apk.Error,
    _sort_baseline.Error,
    _zipalign.Error,
    _zipalignment.Error,
    _zipinfo.Error,
    zipfile.BadZipFile,
)


def main() -> None:
    @click.group(help="""
        repro-apk - scripts to make android apks reproducible
    """)
    @click.version_option(__version__)
    def cli() -> None:
        pass

    @cli.group(help="""
        Parse/dump android binary XML (AXML) or resources (ARSC).
    """)
    def binres() -> None:
        pass

    @binres.command(help="""
        Parse & dump ARSC or AXML.
    """, name="dump")
    @click.option("--apk", metavar="APK", type=click.Path(exists=True, dir_okay=False),
                  help="APK that contains the AXML/ARSC file(s).")
    @click.option("--json", is_flag=True, help="Output JSON.")
    @click.option("--xml", is_flag=True, help="Output XML (AXML only).")
    @click.option("--deref", is_flag=True, help="Follow references (with --xml).")
    @click.option("--prolog", is_flag=True, help="Output XML prolog (with --xml).")
    @click.option("-q", "--quiet", is_flag=True, help="Don't show filenames.")
    @click.option("-v", "--verbose", is_flag=True, help="Be verbose.")
    @click.argument("files_or_patterns", metavar="FILE_OR_PATTERN...", nargs=-1, required=True)
    @click.pass_context
    def binres_dump(ctx: click.Context, files_or_patterns: Tuple[str, ...], apk: str,
                    json: bool, deref: bool, prolog: bool, quiet: bool, xml: bool, verbose: bool) -> None:
        if json and xml:
            raise click.exceptions.BadParameter("Conflicting options: --json and --xml.", ctx)
        if deref and not xml:
            raise click.exceptions.BadParameter("Conflicting options: --deref without --xml.", ctx)
        if prolog and not xml:
            raise click.exceptions.BadParameter("Conflicting options: --prolog without --xml.", ctx)
        if apk:
            _binres.dump_apk(apk, *files_or_patterns, json=json, quiet=quiet,
                             verbose=verbose, xml=xml, xml_deref=deref, xml_prolog=prolog)
        else:
            _binres.dump(*files_or_patterns, json=json, quiet=quiet, verbose=verbose,
                         xml=xml, xml_deref=deref, xml_prolog=prolog)

    @binres.command(help="""
        Quickly get appid & version code/name from APK(s).
    """, name="fastid")
    @click.option("--json", is_flag=True, help="Output JSON.")
    @click.option("--short", is_flag=True, help="Only show values.")
    @click.argument("apks", metavar="APK...", nargs=-1, required=True,
                    type=click.Path(exists=True, dir_okay=False))
    def binres_fastid(apks: Tuple[str, ...], json: bool, short: bool) -> None:
        _binres.fastid(*apks, json=json, short=short)

    @binres.command(help="""
        Quickly get permissions from APK(s).
    """, name="fastperms")
    @click.option("--json", is_flag=True, help="Output JSON.")
    @click.option("--with-id", is_flag=True,
                  help="Also get appid & version code/name.")
    @click.option("-q", "--quiet", is_flag=True, help="Don't show filenames.")
    @click.argument("apks", metavar="APK...", nargs=-1, required=True,
                    type=click.Path(exists=True, dir_okay=False))
    def binres_fastperms(apks: Tuple[str, ...], json: bool, quiet: bool, with_id: bool) -> None:
        _binres.fastperms(*apks, json=json, quiet=quiet, with_id=with_id)

    @binres.command(help="""
        Dump basic manifest info from APK(s) as JSON.
    """, name="manifest-info")
    @click.argument("apks", metavar="APK...", nargs=-1, required=True,
                    type=click.Path(exists=True, dir_okay=False))
    def binres_manifest_info(apks: Tuple[str, ...]) -> None:
        if _binres.manifest_info(*apks):
            sys.exit(1)

    @cli.command(help="""
        Diff ZIP file metadata.
    """)
    @click.option("--no-additional", is_flag=True, help="Skip additional tests.")
    @click.option("--no-lfh-extra", is_flag=True, help="Ignore LFH extra field.")
    @click.option("--no-offsets", is_flag=True, help="Ignore header offsets.")
    @click.option("--no-ordering", is_flag=True, help="Ignore entry ordering.")
    @click.argument("zipfile1", type=click.Path(exists=True, dir_okay=False))
    @click.argument("zipfile2", type=click.Path(exists=True, dir_okay=False))
    def diff_zip_meta(zipfile1: str, zipfile2: str, no_additional: bool, no_lfh_extra: bool,
                      no_offsets: bool, no_ordering: bool) -> None:
        verbosity = _diff_zip_meta.Verbosity(
            additional=not no_additional,
            lfh_extra=not no_lfh_extra,
            offsets=not no_offsets,
            ordering=not no_ordering,
        )
        if _diff_zip_meta.diff_zip_meta(zipfile1, zipfile2, verbosity=verbosity):
            sys.exit(4)

    @cli.command(help="""
        Dump resources.arsc (extracted or inside an APK) using aapt2.
    """)
    @click.option("--apk", is_flag=True,
                  help="ARSC_OR_APK is an APK, not an extracted resources.arsc.")
    @click.argument("arsc_or_apk", type=click.Path(exists=True, dir_okay=False))
    def dump_arsc(arsc_or_apk: str, apk: bool) -> None:
        if apk:
            _dump_arsc.dump_arsc_apk(arsc_or_apk)
        else:
            _dump_arsc.dump_arsc(arsc_or_apk)

    @cli.command(help="""
        Dump Android binary XML (extracted or inside an APK) using aapt2.
    """)
    @click.option("--apk", metavar="APK", type=click.Path(exists=True, dir_okay=False),
                  help="APK that contains the (non-extracted) AXML file.")
    @click.argument("axml", type=click.Path(dir_okay=False))
    def dump_axml(axml: str, apk: str) -> None:
        if apk:
            _dump_axml.dump_axml_apk(apk, axml)
        else:
            _dump_axml.dump_axml(axml)

    @cli.command(help="""
        Dump baseline.prof/baseline.profm (extracted or inside an APK).
    """)
    @click.option("--apk", is_flag=True,
                  help="PROF_OR_APK is an APK, not an extracted .prof/.profm.")
    @click.option("-v", "--verbose", is_flag=True, help="Be verbose.")
    @click.argument("prof_or_apk", type=click.Path(exists=True, dir_okay=False))
    def dump_baseline(prof_or_apk: str, apk: bool, verbose: bool) -> None:
        if apk:
            _dump_baseline.dump_baseline_apk(prof_or_apk, verbose=verbose)
        else:
            _dump_baseline.dump_baseline(prof_or_apk, verbose=verbose)

    @cli.command(help="""
        Recompress with different compression level.

        Specify which files to change by providing at least one fnmatch-style
        PATTERN, e.g. 'assets/foo/*.bar'.
    """)
    @click.option("-v", "--verbose", is_flag=True, help="Be verbose.")
    @click.argument("input_apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("output_apk", type=click.Path(dir_okay=False))
    @click.argument("compresslevel", type=click.INT)
    @click.argument("patterns", metavar="PATTERN...", nargs=-1, required=True)
    def fix_compresslevel(input_apk: str, output_apk: str, compresslevel: int,
                          patterns: Tuple[str, ...], verbose: bool) -> None:
        _fix_compresslevel.fix_compresslevel(input_apk, output_apk, compresslevel,
                                             *patterns, verbose=verbose)

    @cli.command(help="""
        Process ZIP entries using an external command.

        Runs the command for each specified file, providing the old file
        contents as stdin and using stdout as the new file contents.

        The provided command is split on whitespace to allow passing arguments
        (e.g. 'foo --bar'), but shell syntax is not supported.

        Specify which files to process by providing at least one fnmatch-style
        PATTERN, e.g. 'META-INF/services/*'.
    """)
    @click.option("--compresslevel", multiple=True, metavar="PATTERN:LEVELS",
                  help="Specify compresslevel(s) (e.g '*.foo:6,9').")
    @click.option("-v", "--verbose", is_flag=True, help="Be verbose.")
    @click.argument("input_apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("output_apk", type=click.Path(dir_okay=False))
    @click.argument("command")
    @click.argument("patterns", metavar="PATTERN...", nargs=-1, required=True)
    @click.pass_context
    def fix_files(ctx: click.Context, input_apk: str, output_apk: str, command: str,
                  patterns: Tuple[str, ...], compresslevel: Tuple[str, ...],
                  verbose: bool) -> None:
        try:
            clevels = _fix_files.compresslevels_from_spec(*compresslevel)
        except ValueError as e:
            raise click.exceptions.BadParameter(str(e), ctx)
        _fix_files.fix_files(input_apk, output_apk, tuple(command.split()),
                             *patterns, compresslevels=clevels, verbose=verbose)

    @cli.command(help="""
        Change line endings from LF to CRLF (or vice versa).

        Specify which files to change by providing at least one fnmatch-style
        PATTERN, e.g. 'META-INF/services/*'.
    """)
    @click.option("--from-crlf/--to-crlf", is_flag=True,
                  help="Change from CRLF to LF.  [default: LF to CRLF]")
    @click.option("--compresslevel", multiple=True, metavar="PATTERN:LEVELS",
                  help="Specify compresslevel(s) (e.g '*.foo:6,9').")
    @click.option("-v", "--verbose", is_flag=True, help="Be verbose.")
    @click.argument("input_apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("output_apk", type=click.Path(dir_okay=False))
    @click.argument("patterns", metavar="PATTERN...", nargs=-1, required=True)
    @click.pass_context
    def fix_newlines(ctx: click.Context, input_apk: str, output_apk: str,
                     patterns: Tuple[str, ...], from_crlf: bool,
                     compresslevel: Tuple[str, ...], verbose: bool) -> None:
        replace = ("\r\n", "\n") if from_crlf else ("\n", "\r\n")
        try:
            clevels = _fix_newlines.compresslevels_from_spec(*compresslevel)
        except ValueError as e:
            raise click.exceptions.BadParameter(str(e), ctx)
        _fix_newlines.fix_newlines(input_apk, output_apk, *patterns, compresslevels=clevels,
                                   replace=replace, verbose=verbose)

    @cli.command(help="""
        Replace non-deterministic R8 pg-map-id in classes{,N}.dex and update
        checksums, also in baseline.prof.
    """)
    @click.argument("input_dir_or_apk", type=click.Path(exists=True, dir_okay=True))
    @click.argument("output_dir_or_apk", type=click.Path(dir_okay=True))
    @click.argument("pg_map_id")
    def fix_pg_map_id(input_dir_or_apk: str, output_dir_or_apk: str, pg_map_id: str) -> None:
        if os.path.isdir(input_dir_or_apk):
            _fix_pg_map_id.fix_pg_map_id(input_dir_or_apk, output_dir_or_apk, pg_map_id)
        else:
            _fix_pg_map_id.fix_pg_map_id_apk(input_dir_or_apk, output_dir_or_apk, pg_map_id)

    @cli.command(help="""
        List ZIP entries with compression level.

        You can optionally specify which files to list by providing one or more
        fnmatch-style patterns, e.g. 'assets/foo/*.bar'.
    """)
    @click.option("--levels", metavar="LEVELS", help="Level(s) (0-9) to try"
                  f" [default: '{','.join(map(str, _list_compresslevel.LEVELS))}'].")
    @click.argument("apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("patterns", metavar="[PATTERN...]", nargs=-1)
    @click.pass_context
    def list_compresslevel(ctx: click.Context, apk: str, patterns: Tuple[str, ...],
                           levels: Optional[str]) -> None:
        try:
            clevels = _list_compresslevel.levels_from_spec(levels)
        except ValueError as e:
            raise click.exceptions.BadParameter(str(e), ctx)
        _list_compresslevel.list_compresslevel(apk, *patterns, levels=clevels)

    @cli.command(help="""
        Remove entries from ZIP file.

        Specify which files to remove by providing at least one fnmatch-style
        PATTERN, e.g. 'META-INF/MANIFEST.MF'.
    """)
    @click.option("-v", "--verbose", is_flag=True, help="Be verbose.")
    @click.argument("input_apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("output_apk", type=click.Path(dir_okay=False))
    @click.argument("patterns", metavar="PATTERN...", nargs=-1, required=True)
    def rm_files(input_apk: str, output_apk: str, patterns: Tuple[str, ...],
                 verbose: bool) -> None:
        _rm_files.rm_files(input_apk, output_apk, *patterns, verbose=verbose)

    @cli.command(help="""
        Sort (and realign) the ZIP entries of an APK.
    """)
    @click.option("--no-realign", is_flag=True, help="Do not realign.")
    @click.option("--no-force-align", is_flag=True, help="Do not force recreating alignment.")
    @click.option("--reset-lh-extra", is_flag=True, help="Reset ZIP LH extra fields using CD.")
    @click.argument("input_apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("output_apk", type=click.Path(dir_okay=False))
    def sort_apk(input_apk: str, output_apk: str, no_realign: bool,
                 no_force_align: bool, reset_lh_extra: bool) -> None:
        _sort_apk.sort_apk(input_apk, output_apk, realign=not no_realign,
                           force_align=not no_force_align, reset_lh_extra=reset_lh_extra)

    @cli.command(help="""
        Sort baseline.profm (extracted or inside an APK).
    """)
    @click.option("--apk", is_flag=True,
                  help="PROF_OR_APK is an APK, not an extracted .profm.")
    @click.argument("input_prof_or_apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("output_prof_or_apk", type=click.Path(dir_okay=False))
    def sort_baseline(input_prof_or_apk: str, output_prof_or_apk: str, apk: bool) -> None:
        if apk:
            _sort_baseline.sort_baseline_apk(input_prof_or_apk, output_prof_or_apk)
        else:
            _sort_baseline.sort_baseline(input_prof_or_apk, output_prof_or_apk)

    @cli.command(help="""
        Align uncompressed ZIP/APK entries to 4-byte boundaries (and .so shared
        object files to 4096-byte boundaries with -p/--page-align, or other page
        sizes with -P/--page-size).
    """)
    @click.option("-p", "--page-align", is_flag=True,
                  help="Use 4096-byte memory page alignment for .so files.")
    @click.option("-P", "--page-size", type=click.INT, metavar="N",
                  help="Use N*1024-byte memory page alignment for .so files.")
    @click.option("--pad-like-apksigner", is_flag=True,
                  help="Use 0xd935 Android ZIP Alignment Extra Field instead of zero padding.")
    @click.option("--replace", is_flag=True, help="Always replace existing alignment.")
    @click.option("--copy-extra", is_flag=True,
                  help="Copy extra bytes between ZIP entries.")
    @click.option("--no-update-lfh", is_flag=True,
                  help="Don't update the LFH using the data descriptor.")
    @click.argument("align", nargs=-1, metavar="[ALIGN]")
    @click.argument("input_apk", type=click.Path(exists=True, dir_okay=False))
    @click.argument("output_apk", type=click.Path(dir_okay=False))
    @click.pass_context
    def zipalign(ctx: click.Context, align: Tuple[str, ...], input_apk: str,
                 output_apk: str, page_align: bool, page_size: Optional[int],
                 pad_like_apksigner: bool, replace: bool, copy_extra: bool,
                 no_update_lfh: bool) -> None:
        if len(align) > 1:
            s = "s" if len(align) > 2 else ""
            ctx.fail(f"Got unexpected extra argument{s} ({' '.join(align[1:])})")
        if align not in ((), ("4",)):
            raise click.exceptions.BadParameter("ALIGN must be 4.", ctx)
        if page_size not in (None, 4, 16, 64):
            print("Warning: specified page size is not 4, 16, or 64 KiB", file=sys.stderr)
        _zipalign.zipalign(input_apk, output_apk, page_align=bool(page_align or page_size),
                           page_size=page_size, pad_like_apksigner=pad_like_apksigner,
                           replace=replace, copy_extra=copy_extra, update_lfh=not no_update_lfh)

    @cli.command(help="""
        Show info about ZIP alignment.
    """)
    @click.argument("apks", metavar="APK...", nargs=-1, required=True,
                    type=click.Path(exists=True, dir_okay=False))
    def zipalignment(apks: Tuple[str, ...]) -> None:
        _zipalignment.zipalignment(*apks)

    @cli.command(help="""
        List ZIP entries (like zipinfo).

        The -l/--long option adds the compressed size before the compression
        type; -e/--extended does the same, adds the CRC32 checksum before the
        filename as well, uses a more standard date format, and treats filenames
        ending with a "/" as directories.
    """)
    @click.option("-1", "--filenames-only", is_flag=True,
                  help="Only print filenames, one per line.")
    @click.option("-e", "--extended", is_flag=True, help="Use extended output format.")
    @click.option("-l", "--long", is_flag=True, help="Use long output format.")
    @click.option("--sort-by-offset", is_flag=True, help="Sort entries by header offset.")
    @click.argument("zipfile", type=click.Path(exists=True, dir_okay=False))
    def zipinfo(zipfile: str, extended: bool, filenames_only: bool, long: bool,
                sort_by_offset: bool) -> None:
        if filenames_only and not (extended or long):
            _zipinfo.zip_filenames(zipfile, sort_by_offset=sort_by_offset)
        else:
            _zipinfo.zipinfo(zipfile, extended=extended, long=long,
                             sort_by_offset=sort_by_offset)

    try:
        cli(prog_name=NAME)
    except ERRORS as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(3)
    except BrokenPipeError:
        pass


if __name__ == "__main__":
    main()

# vim: set tw=80 sw=4 sts=4 et fdm=marker :



#!/usr/bin/python3
# encoding: utf-8
# SPDX-FileCopyrightText: 2024 FC (Fay) Stegerman <flx@obfusk.net>
# SPDX-License-Identifier: AGPL-3.0-or-later

import struct
import sys
import zipfile
import zlib

from fnmatch import fnmatch
from typing import List, Optional

LEVELS = (9, 6, 4, 1)


class Error(RuntimeError):
    pass


def list_compresslevel(apk: str, *patterns: str, levels: Optional[List[int]] = None) -> None:
    if not levels:
        levels = list(LEVELS)
    with open(apk, "rb") as fh_raw:
        with zipfile.ZipFile(apk) as zf:
            for info in zf.infolist():
                if patterns and not fnmatches_with_negation(info.filename, *patterns):
                    continue
                lvls = []
                if info.compress_type == 8:
                    fh_raw.seek(info.header_offset)
                    n, m = struct.unpack("<HH", fh_raw.read(30)[26:30])
                    fh_raw.seek(info.header_offset + 30 + m + n)
                    ccrc = 0
                    size = info.compress_size
                    while size > 0:
                        ccrc = zlib.crc32(fh_raw.read(min(size, 4096)), ccrc)
                        size -= 4096
                    with zf.open(info) as fh:
                        comps = {lvl: zlib.compressobj(lvl, 8, -15) for lvl in levels}
                        ccrcs = {lvl: 0 for lvl in levels}
                        while True:
                            data = fh.read(4096)
                            if not data:
                                break
                            for lvl in levels:
                                ccrcs[lvl] = zlib.crc32(comps[lvl].compress(data), ccrcs[lvl])
                        for lvl in levels:
                            if ccrc == zlib.crc32(comps[lvl].flush(), ccrcs[lvl]):
                                lvls.append(lvl)
                        if not lvls:
                            raise Error(f"Unable to determine compresslevel for {info.filename!r}")
                elif info.compress_type != 0:
                    raise Error(f"Unsupported compress_type {info.compress_type}")
                result = "|".join(map(str, lvls)) if lvls else None
                print(f"filename={info.filename!r} compresslevel={result}")


def fnmatches_with_negation(filename: str, *patterns: str) -> bool:
    r"""
    Filename matching with shell patterns and negation.

    Checks whether filename matches any of the fnmatch patterns.

    An optional prefix "!" negates the pattern, invalidating a successful match
    by any preceding pattern; use a backslash ("\") in front of the first "!"
    for patterns that begin with a literal "!".

    >>> fnmatches_with_negation("foo.xml", "*", "!*.png")
    True
    >>> fnmatches_with_negation("foo.png", "*", "!*.png")
    False
    >>> fnmatches_with_negation("!foo.png", r"\!*.png")
    True

    """
    matches = False
    for p in patterns:
        if p.startswith("!"):
            if fnmatch(filename, p[1:]):
                matches = False
        else:
            if p.startswith(r"\!"):
                p = p[1:]
            if fnmatch(filename, p):
                matches = True
    return matches


def levels_from_spec(spec: Optional[str]) -> List[int]:
    r"""
    Get compresslevels from spec.

    >>> levels_from_spec("1,4-6,9")
    [1, 4, 5, 6, 9]
    >>> levels_from_spec("1,9-6,4")
    [1, 9, 8, 7, 6, 4]
    >>> try:
    ...     levels_from_spec("1,4-x")
    ... except ValueError as e:
    ...     str(e)
    "Invalid LEVELS spec: '4-x'"
    >>> try:
    ...     levels_from_spec("1,10")
    ... except ValueError as e:
    ...     str(e)
    "Invalid LEVELS spec: '10'"

    """
    if not spec:
        return list(LEVELS)
    levels: List[int] = []
    for subspec in spec.split(","):
        try:
            if "-" in subspec:
                m, n = map(int, subspec.split("-"))
                if not (0 <= m <= 9 and 0 <= n <= 9):
                    raise ValueError("Out of range")
                add = list(range(m, n + 1) if m <= n else range(m, n - 1, -1))
            else:
                n = int(subspec)
                if not (0 <= n <= 9):
                    raise ValueError("Out of range")
                add = [n]
        except ValueError as e:
            raise ValueError(f"Invalid LEVELS spec: {subspec!r}") from e
        levels.extend(add)
    return levels


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(prog="list-compresslevel.py")
    parser.add_argument("--levels")
    parser.add_argument("apk", metavar="APK")
    parser.add_argument("patterns", metavar="PATTERN", nargs="*")
    args = parser.parse_args()
    try:
        clevels = levels_from_spec(args.levels)
    except ValueError as e:
        print(f"Error: {e}.")
        sys.exit(1)
    try:
        list_compresslevel(args.apk, *args.patterns, levels=clevels)
    except BrokenPipeError:
        pass

# vim: set tw=80 sw=4 sts=4 et fdm=marker :



#!/usr/bin/python3
# encoding: utf-8
# SPDX-FileCopyrightText: 2023 FC (Fay) Stegerman <flx@obfusk.net>
# SPDX-License-Identifier: AGPL-3.0-or-later

import calendar
import os
import stat
import struct
import time
import zipfile

from dataclasses import dataclass
from typing import Callable, Optional

# FIXME: support more types, needs test cases
# https://sources.debian.org/src/unzip/6.0-27/zipinfo.c/#L1887
SYS_FAT, SYS_UNX, SYS_NTF = (0, 3, 11)
SYSTEM = {SYS_FAT: "fat", SYS_UNX: "unx", SYS_NTF: "ntf"}

# https://sources.debian.org/src/unzip/6.0-27/zipinfo.c/#L2086
EXE_EXTS = {"com", "exe", "btm", "cmd", "bat"}

# https://sources.debian.org/src/unzip/6.0-27/zipinfo.c/#L1896
COMPRESS_TYPE = {
    zipfile.ZIP_STORED: "stor",
    zipfile.ZIP_DEFLATED: "def",  # DEFLATE_TYPE char is appended
    zipfile.ZIP_BZIP2: "bzp2",
    zipfile.ZIP_LZMA: "lzma",
}

# https://sources.debian.org/src/unzip/6.0-27/zipinfo.c/#L1886
# normal, maximum, fast, superfast
DEFLATE_TYPE = "NXFS"

EXTRA_DATA_INFO = {
    # extra, data descriptor
    (False, False): "-",
    (False, True): "l",
    (True, False): "x",
    (True, True): "X",
}


@dataclass(frozen=True)
class Time:
    """Unix time from extra field (UT or UX)."""
    mtime: int
    atime: Optional[int]
    ctime: Optional[int]


class Error(RuntimeError):
    pass


# FIXME
# https://sources.debian.org/src/unzip/6.0-27/zipinfo.c/#L1097
# https://sources.debian.org/src/zip/3.0-12/zip.h/#L211
def format_info(info: zipfile.ZipInfo, *, extended: bool = False,
                long: bool = False) -> str:
    r"""
    Format ZIP entry info.

    >>> zf = zipfile.ZipFile("test/data/crlf.apk")
    >>> info1 = zf.getinfo("META-INF/")
    >>> info2 = zf.getinfo("resources.arsc")
    >>> info3 = zf.getinfo("LICENSE.GPLv3")
    >>> format_info(info1)
    '-rw----     2.0 fat        0 bx defN 17-May-15 11:25 META-INF/'
    >>> format_info(info1, extended=True)
    'drw----     2.0 fat        0 bx        2 defN 2017-05-15 11:25:18 00000000 META-INF/'
    >>> format_info(info2)
    '-rw----     1.0 fat      896 b- stor 09-Jan-01 00:00 resources.arsc'
    >>> format_info(info3)
    '-rw-------  3.0 unx    35823 t- defN 80-Jan-01 00:00 LICENSE.GPLv3'
    >>> info3.external_attr |= 0o644 << 16
    >>> format_info(info3, long=True)
    '-rw-r--r--  3.0 unx    35823 t-    12289 defN 80-Jan-01 00:00 LICENSE.GPLv3'
    >>> format_info(info3, extended=True)
    '-rw-r--r--  3.0 unx    35823 t-    12289 defN 1980-01-01 00:00:00 cece3b93 LICENSE.GPLv3'
    >>> info4 = zipfile.ZipInfo("foo\n.com")
    >>> info4.file_size = info4.compress_size = 0
    >>> format_info(info4)
    '-rwx---     2.0 unx        0 b- stor 80-Jan-01 00:00 foo^J.com'
    >>> info4.extra = b'UT\x05\x00\x01\x00\x00\x00\x00'
    >>> format_info(info4)
    '-rwx---     2.0 unx        0 bx stor 70-Jan-01 00:00 foo^J.com'
    >>> info4.extra = b'UX\x08\x00\x80h\x9e>|h\x9e>'
    >>> format_info(info4)
    '-rwx---     2.0 unx        0 bx stor 03-Apr-17 08:40 foo^J.com'

    """
    if t := extra_field_time(info.extra):
        date_time = tuple(time.localtime(t.mtime))[:6]
    else:
        date_time = info.date_time
    perm = format_permissions(info)
    if extended and info.filename.endswith("/"):
        perm = "d" + perm[1:]                       # directory
    vers = "{}.{}".format(info.create_version // 10,
                          info.create_version % 10)
    syst = SYSTEM.get(info.create_system, "???")
    xinf = "t" if info.internal_attr == 1 else "b"  # text/binary
    if info.flag_bits & 1:
        xinf = xinf.upper()                         # encrypted
    xinf += EXTRA_DATA_INFO[(bool(info.extra), bool(info.flag_bits & 0x08))]
    comp = COMPRESS_TYPE.get(info.compress_type, "????")
    if info.compress_type == zipfile.ZIP_DEFLATED:
        comp += DEFLATE_TYPE[(info.flag_bits >> 1) & 3]
    if extended:
        dt = "{}-{:02d}-{:02d}".format(*date_time[:3])
        tm = "{:02d}:{:02d}:{:02d}".format(*date_time[3:])
    else:
        dt = "{:02d}-{}-{:02d}".format(
            date_time[0] % 100,
            calendar.month_abbr[date_time[1]] or "000",
            date_time[2]
        )
        tm = "{:02d}:{:02d}".format(*date_time[3:5])
    fields = [f"{perm:<11}", vers, syst, f"{info.file_size:>8}", xinf]
    if long or extended:
        fields.append(f"{info.compress_size:>8}")
    fields += [comp, dt, tm]
    if extended:
        fields.append(f"{info.CRC:08x}")
    fields.append(printable_filename(info.filename))
    return " ".join(fields)


# FIXME
# https://sources.debian.org/src/unzip/6.0-27/zipinfo.c/#L2064
def format_permissions(info: zipfile.ZipInfo) -> str:
    """
    Format ZIP entry Unix or FAT permissions.

    >>> zf = zipfile.ZipFile("test/data/crlf.apk")
    >>> info1 = zf.getinfo("META-INF/")
    >>> info2 = zf.getinfo("resources.arsc")
    >>> info3 = zf.getinfo("LICENSE.GPLv3")
    >>> format_permissions(info1)
    '-rw----'
    >>> format_permissions(info2)
    '-rw----'
    >>> format_permissions(info3)
    '-rw-------'
    >>> info3.external_attr = info3.external_attr | (0o644 << 16)
    >>> format_permissions(info3)
    '-rw-r--r--'
    >>> info4 = zipfile.ZipInfo("foo.com")
    >>> format_permissions(info4)
    '-rwx---'

    """
    hi = info.external_attr >> 16
    if hi and info.create_system in (SYS_UNX, SYS_FAT):
        return stat.filemode(hi)
    exe = os.path.splitext(info.filename)[1][1:].lower() in EXE_EXTS
    xat = info.external_attr & 0xFF
    return "".join((
        'd' if xat & 0x10 else '-',
        'r',
        '-' if xat & 0x01 else 'w',
        'x' if xat & 0x10 or exe else '-',
        'a' if xat & 0x20 else '-',
        'h' if xat & 0x02 else '-',
        's' if xat & 0x04 else '-',
    ))


# https://sources.debian.org/src/zip/3.0-12/zip.h/#L217
# https://sources.debian.org/src/zip/3.0-12/zipfile.c/#L6544
def extra_field_time(extra: bytes, local: bool = False) -> Optional[Time]:
    r"""
    Get unix time from extra field (UT or UX).

    >>> t = extra_field_time(b'UT\x05\x00\x01\x00\x00\x00\x00')
    >>> t
    Time(mtime=0, atime=None, ctime=None)
    >>> tuple(time.localtime(t.mtime))
    (1970, 1, 1, 0, 0, 0, 3, 1, 0)
    >>> t = extra_field_time(b'UT\x05\x00\x01\xda\xe9\xe36')
    >>> t
    Time(mtime=920906202, atime=None, ctime=None)
    >>> tuple(time.localtime(t.mtime))
    (1999, 3, 8, 15, 16, 42, 0, 67, 0)
    >>> t = extra_field_time(b'UX\x08\x00\x80h\x9e>|h\x9e>')
    >>> t
    Time(mtime=1050568828, atime=1050568832, ctime=None)
    >>> tuple(time.localtime(t.mtime))
    (2003, 4, 17, 8, 40, 28, 3, 107, 0)
    >>> tuple(time.localtime(t.atime))
    (2003, 4, 17, 8, 40, 32, 3, 107, 0)
    >>> t = extra_field_time(b'UX\x08\x00\xda\xe9\xe36\xda\xe9\xe36')
    >>> t
    Time(mtime=920906202, atime=920906202, ctime=None)
    >>> tuple(time.localtime(t.mtime))
    (1999, 3, 8, 15, 16, 42, 0, 67, 0)
    >>> tuple(time.localtime(t.atime))
    (1999, 3, 8, 15, 16, 42, 0, 67, 0)

    """
    while len(extra) >= 4:
        hdr_id, size = struct.unpack("<HH", extra[:4])
        if size > len(extra) - 4:
            break
        if hdr_id == 0x5455 and size >= 1:
            flags = extra[4]
            if flags & 0x1 and size >= 5:
                mtime = int.from_bytes(extra[5:9], "little")
                atime = ctime = None
                if local:
                    if flags & 0x2 and size >= 9:
                        atime = int.from_bytes(extra[9:13], "little")
                    if flags & 0x4 and size >= 13:
                        ctime = int.from_bytes(extra[13:17], "little")
                return Time(mtime, atime, ctime)
        elif hdr_id == 0x5855 and size >= 8:
            atime = int.from_bytes(extra[4:8], "little")
            mtime = int.from_bytes(extra[8:12], "little")
            return Time(mtime, atime, None)
        extra = extra[size + 4:]
    return None


def printable_filename(s: str) -> str:
    r"""
    Replace ASCII control characters with caret notation (e.g. ^M, ^J) and other
    non-printable characters with backslash escapes like repr().

    >>> printable_filename("foo bar.baz")
    'foo bar.baz'
    >>> printable_filename("foo\r\n\x7fbar\x82\u0fff.baz")
    'foo^M^J^?bar\\x82\\u0fff.baz'

    """
    t = []
    for c in s:
        if c.isprintable():
            t.append(c)
        elif (i := ord(c)) in range(32):
            t.append("^" + chr(i + 64))
        elif i == 127:
            t.append("^?")
        else:
            t.append(repr(c)[1:-1])
    return "".join(t)


def zipinfo(zip_file: str, *, extended: bool = False, long: bool = False,
            sort_by_offset: bool = False, fmt: Callable[..., str] = format_info) -> None:
    """List ZIP entries, like Info-ZIP's zipinfo(1)."""
    with zipfile.ZipFile(zip_file) as zf:
        if sort_by_offset:
            infos = sorted(zf.infolist(), key=lambda i: i.header_offset)
        else:
            infos = zf.infolist()
        size = os.path.getsize(zip_file)
        tot_u = tot_c = 0
        print(f"Archive:  {printable_filename(zip_file)}")
        print(f"Zip file size: {size} bytes, number of entries: {len(infos)}")
        if infos:
            for info in infos:
                tot_u += info.file_size
                tot_c += info.compress_size
                print(fmt(info, extended=extended, long=long))
                if info.flag_bits & 1:  # encrypted
                    tot_c -= 12         # don't count extra 12 header bytes
            s = "" if len(infos) == 1 else "s"
            r = _cfactor(tot_u, tot_c)
            print(f"{len(infos)} file{s}, {tot_u} bytes uncompressed, "
                  f"{tot_c} bytes compressed:  {r}")
        else:
            print("Empty zipfile.")


# https://sources.debian.org/src/unzip/6.0-27/list.c/#L708
# https://sources.debian.org/src/unzip/6.0-27/zipinfo.c/#L913
def _cfactor(u: int, c: int) -> str:
    """
    Compression factor as reported by Info-ZIP's zipinfo(1).

    >>> _cfactor(0, 0)
    '0.0%'
    >>> _cfactor(400, 300)
    '25.0%'
    >>> _cfactor(300, 400)
    '-33.3%'
    >>> _cfactor(84624808, 25825672)
    '69.5%'
    >>> _cfactor(938, 492)
    '47.5%'
    >>> _cfactor(3070914, 1205461)
    '60.8%'

    """
    if not u:
        r, s = 0, ""
    else:
        f, d = (1, u // 1000) if u > 2000000 else (1000, u)
        if u >= c:
            r, s = (f * (u - c) + (d >> 1)) // d, ""
        else:
            r, s = (f * (c - u) + (d >> 1)) // d, "-"
    return f"{s}{r // 10}.{r % 10}%"


def zip_filenames(zip_file: str, *, sort_by_offset: bool = False) -> None:
    """List ZIP entry filenames, one per line."""
    with zipfile.ZipFile(zip_file) as zf:
        if sort_by_offset:
            infos = sorted(zf.infolist(), key=lambda i: i.header_offset)
        else:
            infos = zf.infolist()
        for info in infos:
            print(printable_filename(info.filename))


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(prog="zipinfo.py")
    parser.add_argument("-1", "--filenames-only", action="store_true",
                        help="only print filenames, one per line")
    parser.add_argument("-e", "--extended", action="store_true",
                        help="use extended output format")
    parser.add_argument("-l", "--long", action="store_true",
                        help="use long output format")
    parser.add_argument("--sort-by-offset", action="store_true",
                        help="sort entries by header offset")
    parser.add_argument("zipfile", metavar="ZIPFILE")
    args = parser.parse_args()
    try:
        if args.filenames_only and not (args.extended or args.long):
            zip_filenames(args.zipfile, sort_by_offset=args.sort_by_offset)
        else:
            zipinfo(args.zipfile, extended=args.extended, long=args.long,
                    sort_by_offset=args.sort_by_offset)
    except BrokenPipeError:
        pass

# vim: set tw=80 sw=4 sts=4 et fdm=marker :



from pathlib import Path
import setuptools

__version__ = "0.3.0"

info = Path(__file__).with_name("README.md").read_text(encoding="utf8")

setuptools.setup(
    name="repro-apk",
    url="https://github.com/obfusk/reproducible-apk-tools",
    description="scripts to make android apks reproducible",
    long_description=info,
    long_description_content_type="text/markdown",
    version=__version__,
    author="FC (Fay) Stegerman",
    author_email="flx@obfusk.net",
    license="AGPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
      # "Programming Language :: Python :: 3.14",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    keywords="android apk reproducible",
    entry_points=dict(
        console_scripts=[
            "repro-apk = repro_apk:main",
            "repro-apk-inplace-fix = repro_apk.inplace_fix:main",
        ]
    ),
    packages=["repro_apk"],
    package_data=dict(repro_apk=["py.typed"]),
    python_requires=">=3.8",
    install_requires=["click>=6.0"],
)



#!/usr/bin/python3
# encoding: utf-8
# SPDX-FileCopyrightText: 2024 FC (Fay) Stegerman <flx@obfusk.net>
# SPDX-License-Identifier: AGPL-3.0-or-later

import hashlib
import os
import re
import struct
import zipfile
import zlib

from binascii import hexlify
from typing import Any, Dict, Match, Tuple

DEX_MAGIC = b"dex\n"
DEX_MAGIC_RE = re.compile(rb"dex\n(\d{3})\x00")

PROF_MAGIC = b"pro\x00"
PROF_010_P = b"010\x00"

CLASSES_DEX_RE = re.compile(r"classes\d*\.dex")
ASSET_PROF = "assets/dexopt/baseline.prof"

PG_MAP_ID_RE = re.compile(rb'(~~R8{"backend":"dex".*?"pg-map-id":")([0-9a-f]{7})(")')

ATTRS = ("compress_type", "create_system", "create_version", "date_time",
         "external_attr", "extract_version", "flag_bits")
LEVELS = (9, 6, 4, 1)


class Error(RuntimeError):
    pass


# FIXME: is there a better alternative?
class ReproducibleZipInfo(zipfile.ZipInfo):
    """Reproducible ZipInfo hack."""

    if "_compresslevel" not in zipfile.ZipInfo.__slots__:       # type: ignore[attr-defined]
        if "compress_level" not in zipfile.ZipInfo.__slots__:   # type: ignore[attr-defined]
            raise Error("zipfile.ZipInfo has no ._compresslevel")

    _compresslevel: int
    _override: Dict[str, Any] = {}

    def __init__(self, zinfo: zipfile.ZipInfo, **override: Any) -> None:
        # pylint: disable=W0231
        if override:
            self._override = {**self._override, **override}
        for k in self.__slots__:
            if hasattr(zinfo, k):
                setattr(self, k, getattr(zinfo, k))

    def __getattribute__(self, name: str) -> Any:
        if name != "_override":
            try:
                return self._override[name]
            except KeyError:
                pass
        return object.__getattribute__(self, name)


def fix_pg_map_id(input_dir: str, output_dir: str, map_id: str) -> None:
    file_data = {}
    for filename in [ASSET_PROF] + sorted(os.listdir(input_dir)):
        if re.fullmatch(CLASSES_DEX_RE, filename) or filename == ASSET_PROF:
            print(f"reading {filename!r}...")
            with open(os.path.join(input_dir, *filename.split("/")), "rb") as fh:
                file_data[filename] = fh.read()
    _fix_pg_map_id(file_data, map_id)
    for filename, data in file_data.items():
        print(f"writing {filename!r}...")
        if "/" in filename:
            os.makedirs(os.path.join(output_dir, *filename.split("/")[:-1]), exist_ok=True)
        with open(os.path.join(output_dir, *filename.split("/")), "wb") as fh:
            fh.write(data)


def fix_pg_map_id_apk(input_apk: str, output_apk: str, map_id: str) -> None:
    with open(input_apk, "rb") as fh_raw:
        with zipfile.ZipFile(input_apk) as zf_in:
            with zipfile.ZipFile(output_apk, "w") as zf_out:
                file_data = {}
                for info in zf_in.infolist():
                    if re.fullmatch(CLASSES_DEX_RE, info.filename) or info.filename == ASSET_PROF:
                        print(f"reading {info.filename!r}...")
                        file_data[info.filename] = zf_in.read(info)
                _fix_pg_map_id(file_data, map_id)
                for info in zf_in.infolist():
                    attrs = {attr: getattr(info, attr) for attr in ATTRS}
                    zinfo = ReproducibleZipInfo(info, **attrs)
                    if info.compress_type == 8:
                        fh_raw.seek(info.header_offset)
                        n, m = struct.unpack("<HH", fh_raw.read(30)[26:30])
                        fh_raw.seek(info.header_offset + 30 + m + n)
                        ccrc = 0
                        size = info.compress_size
                        while size > 0:
                            ccrc = zlib.crc32(fh_raw.read(min(size, 4096)), ccrc)
                            size -= 4096
                        with zf_in.open(info) as fh_in:
                            comps = {lvl: zlib.compressobj(lvl, 8, -15) for lvl in LEVELS}
                            ccrcs = {lvl: 0 for lvl in LEVELS}
                            while True:
                                data = fh_in.read(4096)
                                if not data:
                                    break
                                for lvl in LEVELS:
                                    ccrcs[lvl] = zlib.crc32(comps[lvl].compress(data), ccrcs[lvl])
                            for lvl in LEVELS:
                                if ccrc == zlib.crc32(comps[lvl].flush(), ccrcs[lvl]):
                                    zinfo._compresslevel = lvl
                                    break
                            else:
                                raise Error(f"Unable to determine compresslevel for {info.filename!r}")
                    elif info.compress_type != 0:
                        raise Error(f"Unsupported compress_type {info.compress_type}")
                    if re.fullmatch(CLASSES_DEX_RE, info.filename) or info.filename == ASSET_PROF:
                        print(f"writing {info.filename!r}...")
                        zf_out.writestr(zinfo, file_data[info.filename])
                    else:
                        with zf_in.open(info) as fh_in:
                            with zf_out.open(zinfo, "w") as fh_out:
                                while True:
                                    data = fh_in.read(4096)
                                    if not data:
                                        break
                                    fh_out.write(data)


def _fix_pg_map_id(file_data: Dict[str, bytes], map_id: str) -> None:
    crcs = {}
    for filename in file_data:
        if re.fullmatch(CLASSES_DEX_RE, filename):
            print(f"fixing {filename!r}...")
            data = _fix_dex_id_checksum(file_data[filename], map_id.encode())
            file_data[filename] = data
            crcs[filename] = zlib.crc32(data)
    if ASSET_PROF in file_data:
        print(f"fixing {ASSET_PROF!r}...")
        file_data[ASSET_PROF] = _fix_prof_checksum(file_data[ASSET_PROF], crcs)


def _fix_dex_id_checksum(data: bytes, map_id: bytes) -> bytes:
    def repl(m: Match[bytes]) -> bytes:
        print(f"fixing pg-map-id: {m.group(2)!r} -> {map_id!r}")
        return m.group(1) + map_id + m.group(3)

    magic = data[:8]
    if magic[:4] != DEX_MAGIC or not DEX_MAGIC_RE.fullmatch(magic):
        raise Error(f"Unsupported magic {magic!r}")
    print(f"dex version={int(magic[4:7]):03d}")
    checksum, signature = struct.unpack("<I20s", data[8:32])
    fixed_data = re.sub(PG_MAP_ID_RE, repl, data[32:])
    if fixed_data == data[32:]:
        print("(not modified)")
        return data
    fixed_sig = hashlib.sha1(fixed_data).digest()
    print(f"fixing signature: {hexlify(signature).decode()} -> {hexlify(fixed_sig).decode()}")
    fixed_data = fixed_sig + fixed_data
    fixed_checksum = zlib.adler32(fixed_data)
    print(f"fixing checksum: 0x{checksum:x} -> 0x{fixed_checksum:x}")
    return magic + int.to_bytes(fixed_checksum, 4, "little") + fixed_data


def _fix_prof_checksum(data: bytes, crcs: Dict[str, int]) -> bytes:
    magic, data = _split(data, 4)
    version, data = _split(data, 4)
    if magic == PROF_MAGIC:
        if version == PROF_010_P:
            print("prof version=010 P")
            return PROF_MAGIC + PROF_010_P + _fix_prof_010_p_checksum(data, crcs)
        else:
            raise Error(f"Unsupported prof version {version!r}")
    else:
        raise Error(f"Unsupported magic {magic!r}")


def _fix_prof_010_p_checksum(data: bytes, crcs: Dict[str, int]) -> bytes:
    num_dex_files, uncompressed_data_size, compressed_data_size, data = _unpack("<BII", data)
    dex_data_headers = []
    if len(data) != compressed_data_size:
        raise Error("Compressed data size does not match")
    data = zlib.decompress(data)
    if len(data) != uncompressed_data_size:
        raise Error("Uncompressed data size does not match")
    for i in range(num_dex_files):
        profile_key_size, num_type_ids, hot_method_region_size, \
            dex_checksum, num_method_ids, data = _unpack("<HHIII", data)
        profile_key, data = _split(data, profile_key_size)
        filename = profile_key.decode()
        fixed_checksum = crcs[filename]
        if fixed_checksum != dex_checksum:
            print(f"fixing {filename!r} checksum: 0x{dex_checksum:x} -> 0x{fixed_checksum:x}")
        dex_data_headers.append(struct.pack(
            "<HHIII", profile_key_size, num_type_ids, hot_method_region_size,
            fixed_checksum, num_method_ids) + profile_key)
    fixed_data = b"".join(dex_data_headers) + data
    fixed_cdata = zlib.compress(fixed_data, 1)
    fixed_hdr = struct.pack("<BII", num_dex_files, uncompressed_data_size, len(fixed_cdata))
    return fixed_hdr + fixed_cdata


def _unpack(fmt: str, data: bytes) -> Any:
    assert all(c in "<BHI" for c in fmt)
    size = fmt.count("B") + 2 * fmt.count("H") + 4 * fmt.count("I")
    return struct.unpack(fmt, data[:size]) + (data[size:],)


def _split(data: bytes, size: int) -> Tuple[bytes, bytes]:
    return data[:size], data[size:]


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(prog="fix-pg-map-id.py")
    parser.add_argument("input_dir_or_apk", metavar="INPUT_DIR_OR_APK")
    parser.add_argument("output_dir_or_apk", metavar="OUTPUT_DIR_OR_APK")
    parser.add_argument("pg_map_id", metavar="PG_MAP_ID")
    args = parser.parse_args()
    if os.path.isdir(args.input_dir_or_apk):
        fix_pg_map_id(args.input_dir_or_apk, args.output_dir_or_apk, args.pg_map_id)
    else:
        fix_pg_map_id_apk(args.input_dir_or_apk, args.output_dir_or_apk, args.pg_map_id)

# vim: set tw=80 sw=4 sts=4 et fdm=marker :



#!/usr/bin/python3
# encoding: utf-8
# SPDX-FileCopyrightText: 2024 FC (Fay) Stegerman <flx@obfusk.net>
# SPDX-License-Identifier: AGPL-3.0-or-later

import os
import struct
import sys
import zipfile

from collections import namedtuple
from typing import BinaryIO, Optional

ZipData = namedtuple("ZipData", ("cd_offset", "eocd_offset", "cd_and_eocd"))

DEFAULT_PAGE_SIZE = 4


class Error(RuntimeError):
    pass


def zipalign(input_apk: str, output_apk: str, *, page_align: bool = False,
             page_size: Optional[int] = None, pad_like_apksigner: bool = False,
             replace: bool = False, copy_extra: bool = False, update_lfh: bool = True) -> None:
    with zipfile.ZipFile(input_apk, "r") as zf:
        infos = zf.infolist()
    zdata = zip_data(input_apk)
    offsets = {}
    with open(input_apk, "rb") as fhi, open(output_apk, "w+b") as fho:
        for info in sorted(infos, key=lambda info: info.header_offset):
            off_i = fhi.tell()
            if info.header_offset > off_i:
                extra_bytes = info.header_offset - off_i
                if copy_extra:
                    fho.write(fhi.read(extra_bytes))
                else:
                    fhi.seek(extra_bytes, os.SEEK_CUR)
            hdr = fhi.read(30)
            if hdr[:4] != b"\x50\x4b\x03\x04":
                raise Error("Expected local file header signature")
            n, m = struct.unpack("<HH", hdr[26:30])
            hdr += fhi.read(n + m)
            if info.filename in offsets:
                raise Error(f"Duplicate ZIP entry: {info.filename!r}")
            offsets[info.filename] = off_o = fho.tell()
            if info.compress_type == 0:
                hdr = _align_zip_entry(
                    info, hdr, n, m, off_o, page_align=page_align, page_size=page_size,
                    pad_like_apksigner=pad_like_apksigner, replace=replace)
            if info.flag_bits & 0x08:
                fhi.seek(info.compress_size, os.SEEK_CUR)
                data_descriptor = fhi.read(12)
                if data_descriptor[:4] == b"\x50\x4b\x07\x08":
                    data_descriptor += fhi.read(4)
                fhi.seek(-(info.compress_size + len(data_descriptor)), os.SEEK_CUR)
                if update_lfh:
                    hdr = hdr[:14] + data_descriptor[-12:] + hdr[26:]
            else:
                data_descriptor = b""
            fho.write(hdr)
            _copy_bytes(fhi, fho, info.compress_size + len(data_descriptor))
        extra_bytes = zdata.cd_offset - fhi.tell()
        if copy_extra:
            _copy_bytes(fhi, fho, extra_bytes)
        else:
            fhi.seek(extra_bytes, os.SEEK_CUR)
        cd_offset = fho.tell()
        for info in infos:
            hdr = fhi.read(46)
            if hdr[:4] != b"\x50\x4b\x01\x02":
                raise Error("Expected central directory file header signature")
            n, m, k = struct.unpack("<HHH", hdr[28:34])
            hdr += fhi.read(n + m + k)
            off = int.to_bytes(offsets[info.filename], 4, "little")
            hdr = hdr[:42] + off + hdr[46:]
            fho.write(hdr)
        eocd_offset = fho.tell()
        fho.write(zdata.cd_and_eocd[zdata.eocd_offset - zdata.cd_offset:])
        fho.seek(eocd_offset + 8)
        fho.write(struct.pack("<HHLL", len(offsets), len(offsets),
                              eocd_offset - cd_offset, cd_offset))


# NB: doesn't sync local & CD headers!
def _align_zip_entry(info: zipfile.ZipInfo, hdr: bytes, n: int, m: int, off_o: int, *,
                     page_align: bool = False, page_size: Optional[int] = None,
                     pad_like_apksigner: bool = False, replace: bool = False) -> bytes:
    psize = DEFAULT_PAGE_SIZE if page_size is None else page_size
    align = psize * 1024 if page_align and info.filename.endswith(".so") else 4
    new_off = 30 + n + m + off_o
    old_xtr = hdr[30 + n:30 + n + m]
    new_xtr = b""
    while len(old_xtr) >= 4:
        hdr_id, size = struct.unpack("<HH", old_xtr[:4])
        if size > len(old_xtr) - 4:
            break
        if not (hdr_id == 0 and size == 0):
            if hdr_id == 0xd935:
                if size >= 2:
                    align = int.from_bytes(old_xtr[4:6], "little")
            else:
                new_xtr += old_xtr[:size + 4]
        old_xtr = old_xtr[size + 4:]
    if replace or new_off % align != 0:
        if pad_like_apksigner:
            pad = (align - (new_off - m + len(new_xtr) + 6) % align) % align
            xtr = new_xtr + struct.pack("<HHH", 0xd935, 2 + pad, align) + pad * b"\x00"
        else:
            pad = (align - (new_off - m + len(new_xtr)) % align) % align
            xtr = new_xtr + pad * b"\x00"
        m_b = int.to_bytes(len(xtr), 2, "little")
        hdr = hdr[:28] + m_b + hdr[30:30 + n] + xtr
    return hdr


def _copy_bytes(fhi: BinaryIO, fho: BinaryIO, size: int, blocksize: int = 4096) -> None:
    while size > 0:
        data = fhi.read(min(size, blocksize))
        if not data:
            break
        size -= len(data)
        fho.write(data)
    if size != 0:
        raise Error("Unexpected EOF")


def zip_data(apkfile: str, count: int = 1024) -> ZipData:
    with open(apkfile, "rb") as fh:
        fh.seek(-min(os.path.getsize(apkfile), count), os.SEEK_END)
        data = fh.read()
        pos = data.rfind(b"\x50\x4b\x05\x06")
        if pos == -1:
            raise Error("Expected end of central directory record (EOCD)")
        fh.seek(pos - len(data), os.SEEK_CUR)
        eocd_offset = fh.tell()
        fh.seek(16, os.SEEK_CUR)
        cd_offset = int.from_bytes(fh.read(4), "little")
        fh.seek(cd_offset)
        cd_and_eocd = fh.read()
    return ZipData(cd_offset, eocd_offset, cd_and_eocd)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(prog="zipalign.py")
    parser.add_argument("-p", "--page-align", action="store_true",
                        help="use 4096-byte memory page alignment for .so files")
    parser.add_argument("-P", "--page-size", metavar="N", type=int,
                        help="use N*1024-byte memory page alignment for .so files")
    parser.add_argument("--pad-like-apksigner", action="store_true",
                        help="use 0xd935 Android ZIP Alignment Extra Field "
                             "instead of zero padding")
    parser.add_argument("--replace", action="store_true", help="always replace existing alignment")
    parser.add_argument("--copy-extra", action="store_true",
                        help="copy extra bytes between ZIP entries")
    parser.add_argument("--no-update-lfh", action="store_false", dest="update_lfh",
                        help="don't update the LFH using the data descriptor")
    parser.add_argument("align", metavar="ALIGN", nargs="?", type=int, default=4)
    parser.add_argument("input_apk", metavar="INPUT_APK")
    parser.add_argument("output_apk", metavar="OUTPUT_APK")
    args = parser.parse_args()
    if args.align != 4:
        raise Error("ALIGN must be 4")
    if args.page_size not in (None, 4, 16, 64):
        print("Warning: specified page size is not 4, 16, or 64 KiB", file=sys.stderr)
    zipalign(args.input_apk, args.output_apk, page_align=bool(args.page_align or args.page_size),
             page_size=args.page_size, pad_like_apksigner=args.pad_like_apksigner,
             replace=args.replace, copy_extra=args.copy_extra, update_lfh=args.update_lfh)

# vim: set tw=80 sw=4 sts=4 et fdm=marker :
